import unittest
import numpy as np

from htlogicalgates.stabilizercode import *


class TestStabilizerCode(unittest.TestCase):
    def test_constructor_string(self):
        c = StabilizerCode("4_2_2")
        self.assertEqual(c.n, 4)
        self.assertEqual(c.k, 2)
        np.testing.assert_equal(c.get_e_matrix().T,
                                      np.array([[1, 1, 0, 0, 0, 0, 0, 0],
                                                [1, 0, 1, 0, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 0, 1, 0],
                                                [0, 0, 0, 0, 1, 1, 0, 0],
                                                [1, 1, 1, 1, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 1, 1, 1]]))

    def test_constructor_string_int(self):
        c = StabilizerCode("trivial", 5)
        self.assertEqual(c.n, 5)
        self.assertEqual(c.k, 5)

    def test_constructor_paulis(self):
        x_log = ["X0 X1", "X0 X2"]
        z_log = ["Z0 Z2", "Z0 Z1"]
        stab = ["X0 X1 X2 X3", "Z0 Z1 Z2 Z3"]
        c = StabilizerCode(x_log, z_log, stab)
        self.assertEqual(c.n, 4)
        self.assertEqual(c.k, 2)
        np.testing.assert_equal(c.get_e_matrix().T,
                                      np.array([[1, 1, 0, 0, 0, 0, 0, 0],
                                                [1, 0, 1, 0, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 0, 1, 0],
                                                [0, 0, 0, 0, 1, 1, 0, 0],
                                                [1, 1, 1, 1, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 1, 1, 1]]))

    def test_constructor_checks(self):
        x_log = ["X0 X1 X2", "X0 X2"]
        z_log = ["Z0 Z2", "Z0 Z1"]
        stab = ["X0 X1 X2 X3", "Z0 Z1 Z2 Z3"]
        c = StabilizerCode(x_log, z_log, stab, verify=False)
        self.assertEqual(c.n, 4)
        self.assertEqual(c.k, 2)
        np.testing.assert_equal(c.get_e_matrix().T,
                                      np.array([[1, 1, 1, 0, 0, 0, 0, 0],
                                                [1, 0, 1, 0, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 0, 1, 0],
                                                [0, 0, 0, 0, 1, 1, 0, 0],
                                                [1, 1, 1, 1, 0, 0, 0, 0],
                                                [0, 0, 0, 0, 1, 1, 1, 1]]))
        self.assertRaises(ValueError, lambda: StabilizerCode((x_log, z_log, stab)))

    def test_calculate_distance(self):
        c = StabilizerCode("4_2_2")
        self.assertEqual(c.d, 2)

    def test_stabilizer_generators(self):
        stabs = ["X0 X1", "X0 X1 X2", "X2", "Z0 Z1"]
        self.assertEqual(are_independent_generators(stabs), False)
        new_stabs = reduce_to_stabilizer_generators(stabs)
        self.assertEqual(new_stabs, ["X0 X1 X2", "X2", "Z0 Z1"])
        self.assertEqual(are_independent_generators(new_stabs), True)



import numpy as np
from numpy.typing import NDArray

from .grb_enviroment import *
from .grb_math_interface import *


def create_Perm(n: int, env: Enviroment) -> ExprMatrix:
    perm = ExprMatrix.create_zero(2*n, 2*n)
    one = Expression.create_const(1)
    for i in range(n):
        expr = Expression.create_const(0)
        for j in range(n):
            e = env.create_bin_expression()
            perm[i, j] = e
            perm[i+n, j+n] = e
            expr = expr + e
        env.add_equality_constraint(one, expr, False)
    for i in range(n):
        expr = Expression.create_const(0)
        for j in range(n):
            expr = expr + perm[j, i]
        env.add_equality_constraint(one, expr, False)
    return perm


def create_CZL(con: NDArray, env: Enviroment) -> ExprMatrix:
    n = np.shape(con)[0]
    czl = ExprMatrix.create_identity(2*n)
    assert (len(np.shape(con)) == 2 and n == np.shape(con)[1])
    for i in range(n):
        for j in range(i+1, n):
            if con[i, j] == 1:
                e = env.create_bin_expression()
                czl[i+n, j] = e
                czl[j+n, i] = e
    return czl


def create_SCL(n: int, env: Enviroment) -> ExprMatrix:
    SCL = ExprMatrix.create_zero(2*n, 2*n)
    for i in range(n):
        SCL[i, i] = env.create_bin_expression()
        SCL[i+n, i] = env.create_bin_expression()
        SCL[i, i+n] = env.create_bin_expression()
        SCL[i+n, i+n] = env.create_bin_expression()
        env.add_equality_constraint(Expression.create_const(1),
                                    SCL[i, i]*SCL[i+n, i+n] + SCL[i+n, i]*SCL[i, i+n],
                                    False)
    return SCL


def create_cons_SCL(n: int, env: Enviroment) -> ExprMatrix:
    SCL = ExprMatrix.create_zero(2*n, 2*n)
    for i in range(n):
        SCL[i, i] = env.create_bin_expression()
        SCL[i+n, i] = env.create_bin_expression()
        SCL[i, i+n] = env.create_bin_expression()
        SCL[i+n, i+n] = env.create_bin_expression()
        env.add_equality_constraint(Expression.create_const(1),
                                    SCL[i, i]*SCL[i+n, i+n] + SCL[i+n, i]*SCL[i, i+n],
                                    False)
        env.add_equality_constraint(Expression.create_const(1),
                                    SCL[i+n, i] + SCL[i+n, i+n],
                                    False)
        env.add_equality_constraint(Expression.create_const(1),
                                    SCL[i, i+n] + SCL[i+n, i+n],
                                    False)
    return SCL


def create_reduced_freedom_matrix(n: int, k: int, env: Enviroment) -> ExprMatrix:
    f = ExprMatrix.create_zero(n+k, n+k)
    for i in range(2*k, n+k):
        for j in range(n+k):
            f[i, j] = env.create_bin_expression()
    return f



import unittest

import numpy as np

from htlogicalgates.symplectic_rep.clifford_gate import *
from htlogicalgates.symplectic_rep.helper import *
from htlogicalgates.symplectic_rep.integer_symplectic import symplectic_matrix, symplectic_matrix_inverse, is_symplectic


class TestHelper(unittest.TestCase):
    def test_int_to_array(self):
        i = 34725
        a = int_to_bitarray(i, 21)
        self.assertEqual(len(a), 21)
        b = bitarray_to_int(a)
        self.assertEqual(i, b)
        self.assertRaises(ValueError, lambda: int_to_bitarray(i, 10))

    def test_matrix_rank(self):
        mat = np.array([[1, 1, 1, 1, 0], [1, 0, 0, 0, 0],
                       [0, 0, 1, 1, 1], [0, 1, 0, 0, 1]])
        self.assertEqual(matrix_rank(mat), 3)

    def test_pauli_string_to_list(self):
        arr = pauli_string_to_list("x0 X2 z1 X2 Y3", 4)
        np.testing.assert_array_equal([1, 0, 0, 1, 0, 1, 0, 1], arr)

    def test_AMG_decomposition(self):
        A = np.array([[1, 0, 1, 1],
                      [0, 1, 0, 0],
                      [1, 0, 0, 0],
                      [1, 0, 0, 1]])
        M, G = get_AMG_decomposition(A)
        M_comp = np.array([[1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [1, 0, 1, 0],
                           [1, 0, 1, 1],])
        G_comp = np.array([[0, 0, 0, 0],
                           [0, 0, 0, 0],
                           [0, 0, 0, 0],
                           [0, 0, 0, 0],])
        np.testing.assert_equal(M, M_comp)
        np.testing.assert_equal(G, G_comp)


class TestCliffordGate(unittest.TestCase):
    def test_constructor(self):
        m = np.eye(4, 4, dtype=np.int32)
        p = np.zeros(4, dtype=np.int32)
        me = np.eye(5, 5, dtype=np.int32)
        pe = np.zeros(5, dtype=np.int32)
        pe[-1] = 1
        c0 = Clifford(0, 2)
        c1 = Clifford(0, 0, 2)
        c2 = Clifford(m)
        c3 = Clifford(m, p)
        c4 = Clifford(me, pe, extended=True)
        for c in [c1, c2, c3, c4]:
            self.assertTrue(c0 == c)
        self.assertEqual(c0.id, 0)
        self.assertEqual(c0.pauli_id, 0)
        self.assertTrue(Clifford(32, 2), Clifford(32, 0, 2))
        c5 = Clifford(3432, 45, 3)
        self.assertEqual(c5.num_qubits, 3)
        self.assertEqual(c5.id, 3432)
        self.assertEqual(c5.pauli_id, 45)


    def test_clifford_group_size_mod_pauli(self):
        self.assertEqual(Clifford.group_size_mod_pauli(1), 6)
        self.assertEqual(Clifford.group_size_mod_pauli(2), 720)
        self.assertEqual(Clifford.group_size_mod_pauli(3), 1451520)

class TestRandomSymplectic(unittest.TestCase):
    def test_symplectic_matrix(self):
        for n, i in zip(range(1, 5), [2, 355, 74329, 4234324]):
            m = symplectic_matrix(i, n)
            self.assertTrue(is_symplectic(m))
            self.assertEqual(symplectic_matrix_inverse(m, n), i)



from __future__ import annotations
from typing import overload, Optional

import numpy as np
from numpy.typing import NDArray

from .._utility import _argument_assignment
from .helper import expand_mat_once, get_u_matrix, get_u_bar_matrix, expand_vec_once, int_to_bitarray, bitarray_to_int
from .integer_symplectic import symplectic_matrix, symplectic_matrix_inverse


class Clifford:
    @overload
    def __init__(self, id: int, num_qubits: int):
        """
        Constructs a Clifford gate without Paulis by its integer representation.

        Parameters
        ----------
        id: int 
            Unique clifford id representing the Clifford part modulo Pauli gates.
        num_qubits: int
            Number of qubits the Clifford is defined on.
        """
        pass

    @overload
    def __init__(self, id: int, pauli_id: int, num_qubits: int):
        """
        Constructs a Clifford gate by its integer representation.

        Parameters
        ----------
        id: int
            Unique clifford id representing the Clifford part modulo Pauli gates.
        pauli_id: int
            Integer representing the Pauli part of the Clifford gate.
        num_qubits: int
            Number of qubits the Clifford is defined on.
        """
        pass

    @overload
    def __init__(self, matrix: NDArray):
        """
        Constructs a Clifford gate from a stabilizer tableau.

        Parameters
        ----------
        matrix: NDArray
            Stabilizer tableau of the Clifford operator.
        """
        pass

    @overload
    def __init__(self, matrix: NDArray, phase: NDArray):
        """
        Constructs a Clifford gate from a stabilizer tableau and a phase vector.

        Parameters
        ----------
        matrix: NDArray
            Stabilizer tableau of the Clifford operator.
        phase: NDArray
            Vector holding the phases under transformation of Pauli operators.
        """
        pass

    def __init__(self, *args, **kwargs):
        options = [{"id": int, "pauli_id": int, "num_qubits": int},
                   {"id": int, "num_qubits": int},
                   {"matrix": np.ndarray},
                   {"matrix": np.ndarray, "phase": np.ndarray},
                   {"matrix": np.ndarray, "phase": np.ndarray, "extended": bool}]
        i, a = _argument_assignment(
            options, "Clifford()", *args, **kwargs)
        if i == 0:
            mat = symplectic_matrix(a["id"], a["num_qubits"])
            phase = int_to_bitarray(a["pauli_id"], 2*a["num_qubits"])
        if i == 1:
            mat = symplectic_matrix(a["id"], a["num_qubits"])
            phase = np.zeros(2*a["num_qubits"], dtype=np.int32)
        if i == 2 or i == 3:
            mat = a["matrix"]
            if i == 2:
                phase = np.zeros(len(mat[0]), dtype=np.int32)
            else:
                phase = a["phase"]
        if i == 4:
            mat = a["matrix"]
            phase = a["phase"].astype(np.int32) % 2
        if not a.get("extended", False):
            self._mat = expand_mat_once(mat)
            self._mat[-1, :-1] = _construct_imag_phase(mat).astype(np.int32)
            self._phase = expand_vec_once(phase).astype(np.int32)
        else:
            self._mat = mat.astype(np.int32)
            self._phase = phase.astype(np.int32)
        self._mat %= 2
        self._phase %= 2
        self._N = (len(self._mat) - 1) // 2
        self._M = len(self._phase) - 1

    @property
    def id(self) -> int:
        """
        Returns an integer identifying the pure Clifford part of the Clifford gate.

        Returns
        -------
        int
            Integer representing the Clifford part.
        """
        if 2 * self.num_qubits != self.m:
            raise AttributeError("Must be a valid Clifford element")
        return symplectic_matrix_inverse(self._mat[:-1, :-1], self.num_qubits)

    @property
    def pauli_id(self) -> int:
        """
        Returns an integer identifying the Pauli part of the Clifford gate.

        Returns
        -------
        int
            Integer representing the Pauli part.
        """
        return bitarray_to_int(self._phase[:-1])

    @property
    def symplectic_matrix(self) -> NDArray:
        """
        Returns the symplectic stabilizer tableau of the Clifford gate.

        Returns
        -------
        NDArray
            The stabilizer tableau.
        """
        return self._mat[:-1, :-1]

    @property
    def phase(self) -> NDArray:
        """
        Returns a vector representing the phases of transformed Pauli operators.

        Returns
        -------
        NDArray
            Vector x containing the phases (-1)^x.
        """
        return self._phase[:-1]

    @property
    def num_qubits(self) -> int:
        """
        Returns the number of qubits the Clifford operation is defined on.

        Returns
        -------
        int
            Number of qubits.
        """
        return self._N

    @property
    def m(self) -> int:
        return self._M

    def is_proper(self) -> bool:
        return self.num_qubits == self.m / 2

    def __matmul__(self, other: Clifford) -> Clifford:
        assert self.m == 2 * \
            other.num_qubits, "Cannot multiply Cliffords due to non-matching dimensions"
        phase = other._phase + other._mat.T @ self._phase
        phase += np.diag(other._mat.T @ np.tril(self._mat.T @
                         get_u_bar_matrix(self.num_qubits) @ self._mat, -1) @ other._mat)
        return Clifford((self._mat @ other._mat) % 2, phase % 2, extended=True)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Clifford):
            return False
        return bool(np.all(self.symplectic_matrix == other.symplectic_matrix) and np.all(self.phase == other.phase))

    @staticmethod
    def group_size_mod_pauli(num_qubits: int):
        u = 1
        for i in range(1, num_qubits + 1):
            u *= 4**i - 1
        return 2 ** (num_qubits * num_qubits) * u


def _construct_imag_phase(m: NDArray) -> NDArray:
    return np.diag(m.T @ get_u_matrix(len(m) // 2) @ m) % 2



from typing import List, overload
import numpy as np
from numpy.typing import NDArray

from .resources.resources import load_connectivity
from ._utility import _argument_assignment


class Connectivity:
    @overload
    def __init__(self, name: str):
        """
        Constructs a connectivity object from a known connectivity name. 
        To query available name, call `available_connectivities()`.

        Parameters
        ----------
        name: str 
            Name of the connectivity
        """
        pass

    @overload
    def __init__(self, name: str, num_qubits: int):
        """
        Constructs a connectivity object by its name and number of qubits.
        To query available name, call `available_connectivities()`.

        Parameters
        ----------
        name: str 
            Name of the connectivity.
        num_qubits: int
            Number of qubits of the connectivity.
        """
        pass

    @overload
    def __init__(self, matrix: NDArray):
        """
        Constructs a connectivity object from a matrix. Here, `matrix[i,j] = 1` if qubits `i`
        and `j` are connected and `matrix[i,j] = 0` otherwise.

        Parameters
        ----------
        matrix: NDArray 
            Matrix of the connectivity.
        """
        pass

    def __init__(self, *args, **kwargs):
        options = [{"name": str},
                   {"name": str, "num_qubits": int},
                   {"matrix": np.ndarray}]
        i, a = _argument_assignment(
            options, "Connectivity()", *args, **kwargs)
        if i == 0:
            self._mat = load_connectivity(a["name"], None)
        elif i == 1:
            self._mat = load_connectivity(a["name"], a["num_qubits"])
        elif i == 2:
            self._mat = a["matrix"]
            if len(sh := np.shape(self._mat)) != 2:
                raise ValueError(
                    "Connectivity() matrix input needs to be 2 dimensional")
            if sh[0] != sh[1]:
                raise ValueError(
                    "Connectivity() matrix input needs to be square")
        self._n = np.shape(self._mat)[0]

    @property
    def num_qubits(self) -> int:
        """
        Returns the number of qubits the connectivity is defined on.

        Returns
        ----------
        int
            Number of qubits.
        """
        return self._n

    @property
    def matrix(self) -> NDArray:
        """
        Returns the connectivity matrix. Here, `matrix[i,j] = 1` if qubits `i`
        and `j` are connected and `matrix[i,j] = 0` otherwise.

        Returns
        ----------
        NDArray
            Matrix of the connectivity.
        """
        return self._mat



from __future__ import annotations
from typing import List, Tuple, Iterable
from itertools import product
from copy import deepcopy

import gurobipy as gp
import numpy as np
from numpy.typing import NDArray


class Expression:
    @staticmethod
    def create_var(var: gp.Var) -> Expression:
        return Expression([1], [[var]], 0)

    @staticmethod
    def create_const(c: int) -> Expression:
        return Expression([], [], c)

    def __init__(self, p: List[int], n: List[List[gp.Var]], c: int) -> None:
        self._prefac = p
        self._nums = n
        self._const = c

    def __add__(self, other: Expression):
        assert (isinstance(other, Expression))
        return Expression(self._prefac + other._prefac,
                          self._nums + other._nums,
                          self._const + other._const)

    def __mul__(self, other: Expression):
        assert (isinstance(other, Expression))
        if self.is_zero() or other.is_zero():
            return Expression.create_const(0)
        prefac = [a*b for a, b in product(self._prefac, other._prefac)]
        nums = [a+b for a, b in product(self._nums, other._nums)]
        if self._const != 0:
            prefac += [self._const*a for a in other._prefac]
            nums += other._nums
        if other._const != 0:
            prefac += [other._const*a for a in self._prefac]
            nums += self._nums
        return Expression(prefac, nums, self._const*other._const)

    def __repr__(self) -> str:
        s = ""
        for fac, nums in zip(self._prefac, self._nums):
            s += str(fac) if fac != 1 else ""
            for n in nums:
                s += f"{{{n}}}"
            s += "+"
        if self._const != 0 or len(s) == 0:
            return s + f"{self._const}"
        else:
            return s[:-1]

    def is_const(self) -> bool:
        return len(self._prefac) == 0

    def is_zero(self) -> bool:
        return self.is_const() and self._const == 0

    def copy(self) -> Expression:
        expr: List[List[gp.Var]] = []
        for i in self._nums:
            expr.append([])
            for j in i:
                expr[-1].append(j)
        return Expression(deepcopy(self._prefac), expr, self._const)


class ExprMatrix:
    @staticmethod
    def create_zero(rows: int, columns: int) -> ExprMatrix:
        expr: List[List[Expression]] = []
        for i in range(rows):
            expr.append([])
            for j in range(columns):
                expr[-1].append(Expression.create_const(0))
        return ExprMatrix(expr)

    @staticmethod
    def create_identity(n: int) -> ExprMatrix:
        expr: List[List[Expression]] = []
        for i in range(n):
            expr.append([])
            for j in range(n):
                expr[-1].append(Expression.create_const(1 if i == j else 0))
        return ExprMatrix(expr)

    @staticmethod
    def create_from_array(arr: NDArray) -> ExprMatrix:
        expr: List[List[Expression]] = []
        for i in range(len(arr)):
            expr.append([])
            for j in range(len(arr[0])):
                expr[-1].append(Expression.create_const(arr[i, j]))
        return ExprMatrix(expr)

    def create_expanded_dims(self, rows: int, columns: int, fill_value: Expression = Expression.create_const(0)):
        e = ExprMatrix.create_zero(rows, columns)
        for i in range(rows):
            for j in range(columns):
                if i < self._ROWS and j < self._COLUMNS:
                    e[i, j] = self[i, j].copy()
                else:
                    e[i, j] = fill_value.copy()
        return e

    def transpose(self) -> ExprMatrix:
        e = ExprMatrix.create_zero(self._ROWS, self._COLUMNS)
        for i in range(self._ROWS):
            for j in range(self._COLUMNS):
                e[i, j] = self[j, i]
        return e

    def __init__(self, expr: List[List[Expression]]) -> None:
        self._ROWS = len(expr)
        self._COLUMNS = len(expr[0])
        self._expr = expr

    def __matmul__(self, other: ExprMatrix) -> ExprMatrix:
        assert (self._COLUMNS == other._ROWS)
        expr: List[List[Expression]] = []
        for i in range(self._ROWS):
            expr.append([])
            for j in range(other._COLUMNS):
                expr[-1].append(sum([self[i, k]*other[k, j]
                                for k in range(self._COLUMNS)], start=Expression.create_const(0)))
        return ExprMatrix(expr)

    def __add__(self, other: ExprMatrix) -> ExprMatrix:
        expr: List[List[Expression]] = []
        for i in range(self._ROWS):
            expr.append([])
            for j in range(other._COLUMNS):
                expr[-1].append(self[i, j] + other[i, j])
        return ExprMatrix(expr)

    def __getitem__(self, key: Tuple[int, int]):
        return self._expr[key[0]][key[1]]

    def __setitem__(self, key: Tuple[int, int], val: Expression):
        self._expr[key[0]][key[1]] = val

    def __repr__(self) -> str:
        return self._expr.__repr__()



import json
import os
from typing import Dict, List, Optional, Any

from numpy.typing import NDArray
import numpy as np

from .._utility import ENCODING_FILE, CON_FILE, ENCODING_KEY, DESCR_KEY, CON_KEY

_automated_cons = {"linear": "Connections exist between qubits $i$ and $i+1$.",
                   "circular": "Connections exist between qubits $i$ and $(i+1)%n$.",
                   "all": "Connections exist between all qubits."}

_automated_qeccs = {r"trivial": "The trivial [[n,n,1]] code."}


def get_json_resource(name: str):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    name = os.path.join(script_dir, name)
    with open(name, "r") as f:
        data = json.load(f)
        return data


def get_internal_qeccs() -> Dict:
    return get_json_resource(ENCODING_FILE)


def get_internal_connectivities() -> Dict:
    return get_json_resource(CON_FILE)


def read_external_json(path: str, *loc: Any) -> NDArray:
    with open(path, "r") as f:
        data = json.load(f)
        for l in loc:
            data = data[l]
        return np.array(data, dtype=np.int32)


def load_stabilizercode(name: str, num_qubits: Optional[int] = None) -> NDArray:
    if name in get_internal_qeccs().keys():
        return np.array(get_internal_qeccs()[name][ENCODING_KEY], dtype=np.int32).T
    if num_qubits == None:
        raise ValueError("Please pass a qubit count 'n'!")
    if "trivial" in name:
        return np.eye(2*num_qubits, dtype=np.int32)
    raise ValueError(f"No code found under name '{str(name)}'.")


def load_connectivity(name: str, num_qubits: Optional[int] = None) -> NDArray:
    if name in get_internal_connectivities().keys():
        return np.array(get_internal_connectivities()[name][CON_KEY], dtype=np.int32)
    if num_qubits == None:
        raise ValueError("Please pass a qubit count 'n'")
    elif name in _automated_cons.keys():
        if name in ["all-to-all", "all"]:
            return np.full((num_qubits, num_qubits), 1, dtype=np.int32) - np.identity(num_qubits, dtype=np.int32)
        if name in ["circular", "circle", "circ"]:
            return np.roll(np.identity(num_qubits, dtype=np.int32), shift=1, axis=0) +\
                np.roll(np.identity(num_qubits, dtype=np.int32), shift=-1, axis=0)
        if name in ["linear", "line"]:
            return np.eye(num_qubits, num_qubits, 1, dtype=np.int32) + np.eye(num_qubits, num_qubits, -1, dtype=np.int32)
    raise ValueError(f"No connectivity found under name '{str(name)}'.")


def available_stabilizercodes() -> Dict:
    """
    Returns all precompiled stabilizers codes as a dictionary.

    Returns
    ----------
    Dict
        Dictionary of all precompiled stabilizers codes. Keys are the names of the
        the stabilizers codes and values are explanations.
    """
    return {key: val[DESCR_KEY] for key, val in get_internal_qeccs().items()} | _automated_qeccs


def available_connectivities() -> Dict:
    """
    Returns all precompiled connectivies as a dictionary.

    Returns
    ----------
    Dict
        Dictionary of all precompiled connectivites. Keys are the names of the
        the connectivities and values are explanations.
    """
    return {key: val[DESCR_KEY] for key, val in get_internal_connectivities().items()} | _automated_cons



import unittest

from htlogicalgates.circuit import *


class TestCircuit(unittest.TestCase):
    def test_constructor(self):
        c = Circuit(2)
        self.assertEqual(c.num_qubits, 2)
        c = Circuit("H 2")
        self.assertEqual(c.num_qubits, 3)

    def test_clifford_to_circuit(self):
        for c_id, p_id, n in [(701, 23, 3), (800923, 421, 5), (32112321, 5344, 7)]:
            cliff = Clifford(c_id, p_id, n)
            circ = Circuit(cliff)
            cliff_compare = circ.to_clifford()
            self.assertTrue(cliff == cliff_compare)

    def test_empty_lines_string(self):
        c = Circuit("H 0\n\nH 1")
        self.assertEqual(c.gate_count(), 2)

    def test_comment_string(self):
        c = Circuit("H 0\nCZ 0 1 #Comment\nH 1")
        self.assertEqual(c.gate_count(), 3)

    def test_no_whitespace_string(self):
        c = Circuit("H 0\nH1\nCZ 0 1")
        self.assertEqual(c.gate_count(), 3)

    def test_shallow_optimize(self):
        c = Circuit("H 2\nS 2\nCZ 0 2")
        self.assertEqual(c.gate_count(), 3)
        cliff = c.to_clifford()
        c.shallow_optimize()
        self.assertLessEqual(c.gate_count(), 2)
        self.assertTrue(cliff == c.to_clifford())

    def test_shallow_optimize_pauli(self):
        c = Circuit("H 0\nZ 0")
        c.shallow_optimize()
        self.assertTrue(c.to_clifford() == Circuit("X 0\nH 0").to_clifford())


from functools import cache
import numpy as np
from numpy.typing import NDArray
import copy
from typing import Tuple, List


def matrix_rank(A: NDArray) -> int:
    """ 
    Computes the rank of a matrix A with coefficients in GF(2).
    Taken with permission from package `htstabilizer`
    """
    if (rows := np.shape(A)[0]) != (columns := np.shape(A)[1]):
        nA = np.zeros([max(rows, columns)]*2, dtype=A.dtype)
        nA[0:rows, 0:columns] = A
    else:
        nA = A

    def rref(A: np.ndarray) -> Tuple[np.ndarray, List[int]]:
        """
        Computes the reduced-row echelon form of a matrix A 
        with entries in the binary field GF(2). 

        Parameters
        ----------
        A : np.ndarray
            Input matrix, should only contain elements 0 or 1 as integers. 

        Returns
        -------
        Tuple[np.ndarray, List[int]]
            RREF of input matrix A as well as column indices of the pivot elements. 
        """
        m, n = A.shape  # m: rows, n: cols
        A = A % 2
        pivot_cols = []
        h = 0
        k = 0
        while h < m and k < n:
            found = False
            i = h
            while not found and i < m:
                if A[i, k] == 1:
                    found = True
                    break
                i += 1
            if not found:
                k += 1
            else:
                pivot_cols.append(k)
                temp = copy.deepcopy(A[h, :])
                A[h, :] = A[i, :]
                A[i, :] = temp
                for i in list(range(h)) + list(range(h+1, m)):
                    A[i, :] = (A[i, :] + A[i, k]*A[h, :]) % 2
                h += 1
                k += 1
        return (A, pivot_cols)

    return len(rref(nA)[1])


def pauli_string_to_list(s: str, n: int) -> List[int]:
    out = [0] * (2*n)
    for p in s.split():
        try:
            tar = int(p[1:])
        except ValueError:
            raise ValueError(
                f"Pauli string '{str(p[1:])}' could not be converted to Pauli! Unknown symbol '{p}'")
        if tar >= n:
            raise ValueError(
                f"Pauli string '{str(p[1:])}' could not be converted to Pauli! Index '{str(tar)}' to large for n={str(n)}.")
        if p[0].upper() == "X":
            out[tar] += 1
        elif p[0].upper() == "Y":
            out[tar] += 1
            out[tar+n] += 1
        elif p[0].upper() == "Z":
            out[tar+n] += 1
        else:
            raise ValueError(
                f"Pauli string '{str(p[1:])}' could not be converted to Pauli! Unknown symbol '{p}'")
    return [i % 2 for i in out]


def max_index_of_pauli(s: str) -> int:
    m = -1
    for p in s.split():
        try:
            tar = int(p[1:])
        except ValueError:
            raise ValueError(
                f"Pauli string '{str(p[1:])}' could not be converted to Pauli! Unknown symbol '{p}'")
        m = max(m, tar)
    return m + 1


def int_to_bitarray(i: int, l: int) -> NDArray:
    bitstring = (r"{0:"+str(l)+r"b}").format(i)
    if l < len(bitstring):
        raise ValueError(
            f"Integer '{str(i)}' too large for bitstring of length '{str(l)}'")
    return np.array([1 if j == "1" else 0 for j in bitstring], dtype=np.int32)


def bitarray_to_int(bitarray: NDArray) -> int:
    string = ""
    for i in bitarray:
        string += "1" if i == 1 else "0"
    return int(string, 2)


def expand_mat_once(m: NDArray) -> NDArray:
    rows, columns = np.shape(m)
    m_bar = np.zeros((rows+1, columns+1), dtype=np.int32)
    m_bar[:rows, :columns] = m
    m_bar[-1, -1] = 1
    return m_bar


def expand_vec_once(v: NDArray) -> NDArray:
    d = len(v)
    v_bar = np.zeros((d+1,), dtype=np.int32)
    v_bar[:d] = v
    return v_bar


@cache
def get_u_matrix(n: int) -> NDArray:
    u = np.zeros((2*n, 2*n))
    u[:n, n:2*n] = np.identity(n, dtype=np.int32)
    return u


@cache
def get_omega_matrix(n: int) -> NDArray:
    u = get_u_matrix(n)
    return u + u.T


@cache
def get_u_bar_matrix(n: int) -> NDArray:
    return expand_mat_once(get_u_matrix(n))


def get_AMG_decomposition(A) -> Tuple[NDArray, NDArray]:
    # from <https://arxiv.org/pdf/quant-ph/0406196> Lemma 7
    n = len(A)
    M = np.eye(n, n, dtype=np.int32)
    G = np.zeros((n, n), dtype=np.int32)
    for i in range(n):
        for j in range(i):
            M[i, j] = (A[i, j] + np.dot(M[i, :j], M[j, :j]))%2
    return M, (M@M.T + A)%2


class LinSolver:
    @staticmethod
    def _get_row_adder(control: int, target: int, N: int):
        q = np.identity(N, dtype=np.int32)
        q[target, control] = 1
        return q

    @staticmethod
    def _get_row_swapper(i: int, j: int, N: int):
        q = np.identity(N, dtype=np.int32)
        q[i, i] = q[j, j] = 0
        q[i, j] = q[j, i] = 1
        return q

    def __init__(self, A: NDArray):
        self.N, self.M = np.shape(A)
        self.traf = np.identity(self.N, dtype=np.int32)
        self.ids = []
        A = A.copy()
        row = 0
        for c in range(self.M):
            i = row
            while A[i, c] != 1:
                i += 1
                if i >= self.N:
                    i -= 1
                    break
            A = (LinSolver._get_row_swapper(i, row, self.N)@A) % 2
            self.traf = (LinSolver._get_row_swapper(
                i, row, self.N)@self.traf) % 2
            if A[row, c] == 1:
                for j in range(row+1, self.N):
                    if A[j, c] == 1:
                        A = (LinSolver._get_row_adder(row, j, self.N)@A) % 2
                        self.traf = (LinSolver._get_row_adder(
                            row, j, self.N)@self.traf) % 2
                self.ids.append(c)
                row += 1
            if row >= self.N:
                break
        for i in reversed(range(self.N)):
            c = self.ids[i]
            for j in range(i):
                if A[j, c] == 1:
                    A = (LinSolver._get_row_adder(i, j, self.N)@A) % 2
                    self.traf = (LinSolver._get_row_adder(
                        i, j, self.N)@self.traf) % 2

    def get_solution(self, b: NDArray) -> NDArray:
        q = (self.traf@b) % 2
        s = np.zeros((self.M,), dtype=np.int32)
        for i, e in enumerate(self.ids):
            s[e] = q[i]
        return s



from __future__ import annotations

from typing import Dict, Any

import gurobipy as gp
from gurobipy import GRB
import numpy as np
from numpy.typing import NDArray

from .grb_math_interface import *


class Enviroment:
    NAME_BIN = "bin"
    NAME_CONST = "cst"
    NAME_CONST_CONSTR = "constraint"

    NAME_AND = "and"

    #### Creating model ####

    def __init__(self, log_to_console: bool = False, log_file: str = "",
                 gurobi: Dict[Any, Any] = {}) -> None:
        self._bin_counter = 0
        self._predef_counter = 0

        self.enviroment = gp.Env(logfilename=log_file, empty=True)

        self.enviroment.setParam("LogToConsole", log_to_console)
        self.enviroment.start()
        self.model = gp.Model(env=self.enviroment)
        self.model.setParam(GRB.Param.Presolve, 2)
        self.model.setParam(GRB.Param.Cuts, 2)
        for key, val in gurobi.items():
            self.model.setParam(key, val)

    def set_time_limit(self, time_limit: float):
        self.model.setParam(GRB.Param.TimeLimit, time_limit)

    def create_bin_var(self) -> gp.Var:
        self._bin_counter += 1
        v = self.model.addVar(
            vtype=GRB.BINARY, name=Enviroment.NAME_BIN + str(self._bin_counter))
        return v

    def create_bin_expression(self) -> Expression:
        return Expression.create_var(self.create_bin_var())

    def create_predef_bin_expression(self) -> Tuple[Expression, int]:
        i = self.create_predef_var()
        return Expression.create_var(i), -self._predef_counter

    def create_predef_bin_matrix(self, rows: int, columns: int) -> Tuple[ExprMatrix, NDArray]:
        ids: List[List[int]] = []
        expr: List[List[Expression]] = []
        for i in range(rows):
            ids.append([])
            expr.append([])
            for j in range(columns):
                a, b = self.create_predef_bin_expression()
                expr[-1].append(a)
                ids[-1].append(b)
        return ExprMatrix(expr), np.array(ids, dtype=np.int32)

    def create_predef_var(self) -> gp.Var:
        self._predef_counter += 1
        v = self.model.addVar(
            vtype=GRB.BINARY, name=Enviroment.NAME_CONST + str(self._predef_counter))
        return v

    def set_predef_var(self, value: int, id: int):
        c = self.model.getConstrByName(Enviroment.NAME_CONST_CONSTR + str(id))
        if c is not None:
            self.model.remove(c)
        v = self._get_var_by_id(id)
        self.model.addConstr(
            int(value) == v, name=Enviroment.NAME_CONST_CONSTR + str(id))

    def set_many_predef_var(self, values: NDArray, ids: NDArray):
        self.model.update()
        for i in range(len(values)):
            for j in range(len(values[0])):
                self.set_predef_var(values[i, j], ids[i, j])

    def add_equality_constraint_mat(self, lhs: ExprMatrix, rhs: ExprMatrix, add_slack: bool = True):
        assert (lhs._COLUMNS == rhs._COLUMNS and lhs._ROWS == rhs._ROWS)
        for i in range(lhs._ROWS):
            for j in range(lhs._COLUMNS):
                self.add_equality_constraint(lhs[i, j], rhs[i, j], add_slack)

    def add_equality_constraint(self, lhs: Expression, rhs: Expression, add_slack: bool = True):
        if add_slack:
            slack = self._get_slack_var()
            self.model.addConstr(sum([a*self._add_AND_constraint(b) for a, b in zip(lhs._prefac, lhs._nums)], start=lhs._const) == 2*slack +
                                 sum([a*self._add_AND_constraint(b) for a, b in zip(rhs._prefac, rhs._nums)], start=rhs._const))
        else:
            self.model.addConstr(sum([a*self._add_AND_constraint(b) for a, b in zip(lhs._prefac, lhs._nums)], start=lhs._const) ==
                                 sum([a*self._add_AND_constraint(b) for a, b in zip(rhs._prefac, rhs._nums)], start=rhs._const))

    def set_target_function(self, expr: Expression):
        self.model.setObjective(sum([a*self._add_AND_constraint(b)
                                for a, b in zip(expr._prefac, expr._nums)], start=expr._const), GRB.MINIMIZE)

    def get_model(self) -> gp.Model:
        return self.model

    def solve(self):
        self.model.optimize()

    #### After solving ####

    def evaluate_expression(self, expr: Expression) -> int:
        return expr._const + sum([a*np.prod([round(i.X) for i in b]) for a, b in zip(expr._prefac, expr._nums)])

    def evaluate_matrix(self, mat: ExprMatrix) -> NDArray:
        return np.array([[self.evaluate_expression(j) for j in i] for i in mat._expr], dtype=np.int32)

    def has_solution(self) -> bool:
        return self.model.SolCount > 0

    def get_status(self) -> str:
        if self.has_solution():
            if self.model.Status == GRB.OPTIMAL:
                return "Optimal"
            else:
                return "Bound " + str(self.model.ObjBound)
        else:
            if self.model.Status == GRB.INFEASIBLE:
                return "Infeasible"
            else:
                return "Time out"

    def get_runtime(self) -> float:
        return self.model.Runtime

    def get_work(self) -> float:
        return self.model.Runtime

    #### Internal functions ####

    def _add_AND_constraint(self, vars: List[gp.Var]) -> gp.Var:
        if len(vars) == 1:
            return vars[0]
        if len(vars) == 0:
            raise ValueError("List of variables has size zero!")
        v = self.create_bin_var()
        self.model.addGenConstrAnd(
            v, vars, name=Enviroment.NAME_AND + str(self._bin_counter))
        return v

    def _get_var_by_id(self, id: int) -> gp.Var:
        try:
            if id > 0:
                v = self.model.getVarByName(Enviroment.NAME_BIN + str(id))
            elif id < 0:
                v = self.model.getVarByName(Enviroment.NAME_CONST + str(-id))
        except gp.GurobiError:
            self.model.update()
            return self._get_var_by_id(id)
        if v is None:
            self.model.update()
            return self._get_var_by_id(id)
        return v

    def _get_slack_var(self) -> gp.Var:
        return self.model.addVar(-float("inf"), float("inf"), vtype=GRB.INTEGER)



import numpy as np
from numpy.typing import NDArray
from typing import Tuple, List, overload
from copy import deepcopy
from itertools import chain, combinations
from functools import cached_property
from qsalto import M as MacWilliams

from .symplectic_rep.helper import pauli_string_to_list, max_index_of_pauli
from .resources.resources import load_stabilizercode
from .symplectic_rep.helper import matrix_rank
from ._utility import _argument_assignment


class StabilizerCode:
    @overload
    def __init__(self, name: str):
        """
        Constructs a stabilizer code from a known code name. To query available names, use
        `available_stabilizercodes()`.

        Parameters
        ----------
        name: str 
            Name of the stabilizer code
        """
        pass

    @overload
    def __init__(self, name: str, num_qubits: int):
        """
        Constructs a stabilizer code from a known code name and the number of qubits.
        To query available names, use `available_stabilizercodes()`.

        Parameters
        ----------
        name: str 
            Name of the stabilizer code
        num_qubits: int
            Number of qubits of the stabilizer code
        """
        pass

    @overload
    def __init__(self, paulis: Tuple[List[str], List[str],
                 List[str]], verify: bool = True):
        """
        Constructs a stabilizer code from a set of stabilizer generators and
        logical Pauli operators.

        Parameters
        ----------
        paulis: Tuple[List[str], List[str], List[str]]
            A tuple (x_logicals, z_logicals, stabilizers), representing the logical
            Pauli-X operators, logical Pauli-Z operators, and stabilizers, respectively.
        verify: bool, optional
            Whether to verify commutation relations of the operators, by default True.

        Examples
        ----------

        >>> x_log = ["X0 X1", "X0 X2"]
        >>> z_log = ["Z0 Z2", "Z0 Z1"]
        >>> stab = ["X0 X1 X2 X3", "Z0 Z1 Z2 Z3"]
        >>> c = StabilizerCode((x_log, z_log, stab))
        """

    @overload
    def __init__(self, x_logicals: List[str], z_logicals: List[str],
                 stabilizers: List[str], verify: bool = True):
        """
        Constructs a stabilizer code from a set of stabilizer generators
        and logical Pauli operators.

        Parameters
        ----------
        x_logicals: List[str]
            Logical Pauli-X operators of the code.
        z_logicals: List[str]
            Logical Pauli-Z operators of the code.
        stabilizers: List[str]
            Stabilizers of the code
        verify: bool, optional
            Whether to verify commutation relations of the operators, by default True.

        Examples
        ----------

        >>> x_log = ["X0 X1", "X0 X2"]
        >>> z_log = ["Z0 Z2", "Z0 Z1"]
        >>> stab = ["X0 X1 X2 X3", "Z0 Z1 Z2 Z3"]
        >>> c = StabilizerCode(x_log, z_log, stab)
        """

    @overload
    def __init__(self, truncated_encoding: NDArray):
        """
        Constructs a stabilizer code from a truncated encoding matrix. The columns
        0 to k-1 specifiy the logical Pauli-X operators, the columns k to 2k-1 specify the
        logical Pauli-Z operators. The columns 2k to n+k specify the stabilizers.

        Parameters
        ----------
        truncated_encoding: NDArray
            Truncated encoding matrix of the code.
        """
        pass

    def __init__(self, *args, **kwargs):
        options = [{"name": str},
                   {"name": str, "num_qubits": int},
                   {"paulis": Tuple},
                   {"paulis": Tuple, "verify": bool},
                   {"x_logicals": List, "z_logicals": List, "stabilizers": List},
                   {"x_logicals": List, "z_logicals": List,
                       "stabilizers": List, "verify": bool},
                   {"truncated_encoding": np.ndarray}]
        i, a = _argument_assignment(
            options, "StabilizerCode()", *args, **kwargs)

        def _get_qecc_e_from_paulis(x_logicals, z_logicals, stabilizers, verify) -> NDArray:
            k = len(x_logicals)
            n = len(stabilizers) + k
            if verify:
                if k != len(z_logicals):
                    raise ValueError(
                        "Different number of logical Pauli-X and Pauli-Z operators")
                if n != (m := max([max_index_of_pauli(i) for i in x_logicals + z_logicals + stabilizers])):
                    raise ValueError(
                        f"Wrong number of stabilizer generators for code defined on '{m}' qubits")
            els = [pauli_string_to_list(i, n)
                   for i in x_logicals + z_logicals + stabilizers]
            return np.array(els, dtype=np.int32).T

        if i == 0:
            self._e_mat = load_stabilizercode(a["name"])
        elif i == 1:
            self._e_mat = load_stabilizercode(a["name"], a["num_qubits"])
        elif i == 2:
            if len(a["paulis"]) != 3:
                raise ValueError(
                    f"StabilizerCode() argument got invalid value '{str(a['paulis'])}'")
            self._e_mat = _get_qecc_e_from_paulis(
                a["paulis"][0], a["paulis"][1], a["paulis"][2], True)
            self._check_validity()
        elif i == 3:
            self._e_mat = _get_qecc_e_from_paulis(
                a["paulis"][0], a["paulis"][1], a["paulis"][2], a["verify"])
            if a["verify"]:
                self._check_validity()
        elif i == 4:
            self._e_mat = _get_qecc_e_from_paulis(
                a["x_logicals"], a["z_logicals"], a["stabilizers"], True)
            self._check_validity()
        elif i == 5:
            self._e_mat = _get_qecc_e_from_paulis(
                a["x_logicals"], a["z_logicals"], a["stabilizers"], a["verify"])
            if a["verify"]:
                self._check_validity()
        elif i == 6:
            self._e_mat = a["truncated_encoding"]
            self._check_validity()
        self._distance = -1

    def _check_validity(self):
        if matrix_rank(self.get_e_matrix()[:, 2*self.k:]) != self.n - self.k:
            raise ValueError(
                "Given stabilizer generators are not independent. Get an independent set by calling `reduce_to_stabilizer_generators`")
        e_xz = self.get_e_matrix()
        e_zx = np.roll(self.get_e_matrix(), shift=self.n, axis=0)
        if np.count_nonzero((e_xz[:, 2*self.k:].T@e_zx[:, 2*self.k:]) % 2) != 0:
            raise ValueError("Given stabilizer generators do not commute")
        if np.count_nonzero((e_xz[:, self.k:2*self.k].T@e_zx[:, self.k:2*self.k]) % 2) != 0:
            raise ValueError("Logical Pauli-Z do not commute with themselves")
        if np.count_nonzero((e_xz[:, self.k:].T@e_zx[:, self.k:]) % 2) != 0:
            raise ValueError("Logical Pauli-Z do not commute with stabilizers")
        if np.count_nonzero((e_xz[:, :self.k].T@e_zx[:, :self.k]) % 2) != 0:
            raise ValueError("Logical Pauli-X do not commute with themselves")
        if np.count_nonzero((np.delete(e_xz, range(self.k, 2*self.k), axis=1).T@np.delete(e_zx, range(self.k, 2*self.k), axis=1)) % 2) != 0:
            raise ValueError("Logical Pauli-X do not commute with themselves")
        if np.count_nonzero((e_xz[:, self.k:2*self.k].T@e_zx[:, :self.k]) % 2 - np.eye(self.k, dtype=np.int32)) != 0:
            raise ValueError(
                "Logical Pauli operators do not (anti-)commute correctly")

    def get_e_matrix(self) -> NDArray:
        """
        Returns the truncated stabilizer tableau of the code.

        Returns
        ----------
        NDArray
            Truncated stabilizer tableau.
        """
        return self._e_mat

    @property
    def n(self) -> int:
        """
        Returns the number of physical qubits of the stabilizer code.

        Returns
        ----------
        int
            Number of physical qubits.
        """
        return np.shape(self._e_mat)[0]//2

    @property
    def k(self) -> int:
        """
        Returns the number of logical qubits of the stabilizer code.

        Returns
        ----------
        int
            Number of logical qubits.
        """
        return np.shape(self._e_mat)[1] - self.n

    @cached_property
    def d(self) -> int:
        """
        Returns the distance of the stabilizer code. The process of computing the
        distance may take some time.

        Returns
        ----------
        int
            Distance of the stabilizer code.
        """
        sle_a = np.zeros(self.n + 1, dtype=np.int32)
        for stabs in chain.from_iterable(combinations(self.get_e_matrix()[:, 2*self.k:].T, r)
                                         for r in range(self.n - self.k + 1)):
            if len(stabs) == 0:
                sle_a[0] += 1
                continue
            stab = np.sum(stabs, axis=0) % 2
            sle_a[np.count_nonzero(stab[:self.n] + stab[self.n])] += 1
        sle_b = MacWilliams(self.n) @ sle_a
        return int(np.min(np.nonzero(np.round(sle_b*2**self.k).astype(sle_a.dtype) - sle_a)))

    @property
    def nkd(self) -> Tuple[int, int, int]:
        """
        Returns the number of physical qubits n, the number of logical qubits k,
        and the distance d of the stabilizer code. The process of computing the
        distance may take some time.

        Returns
        ----------
        Tuple[int, int, int]
            The numbers [[n,k,d]] as a tuple.
        """
        return (self.n, self.k, self.d)


def reduce_to_stabilizer_generators(stabilizers: List[str]) -> List[str]:
    """
    Reduces a list of stabilizers to a set of independent generators.

    Parameters
    ----------
    stabilizers
        List of stabilizers to reduce.

    Examples
    ----------
    >>> stabilizers = ["X0 X1 X2 X3", "Z0 Z1 Z2 Z3", "Y0 Y1 Y2 Y3"]
    >>> new_stabilizers = reduce_to_stabilizer_generators(stabilizers)
    >>> print(new_stabilizers)
    ['Z0 Z1 Z2 Z3', 'Y0 Y1 Y2 Y3']
    """
    num_qubits = max([max_index_of_pauli(i) for i in stabilizers])
    stabs = np.array([pauli_string_to_list(i, num_qubits)
                      for i in stabilizers], dtype=np.int32)
    rank = matrix_rank(stabs)
    index = []
    num = np.shape(stabs)[0]
    i = 0
    while i < num - len(index):
        if rank == matrix_rank(a := np.delete(stabs, i, axis=0)):
            stabs = a
            index.append(i)
        else:
            i += 1
    new_stabs = deepcopy(stabilizers)
    for j in index:
        new_stabs.pop(j)
    return new_stabs


def are_independent_generators(stabilizers: List[str]) -> bool:
    return stabilizers == reduce_to_stabilizer_generators(stabilizers)



from .htlogicalgates import tailor_multiple_logical_gates, tailor_logical_gate
from .htlogicalgates import save_results_dictionary, load_results_dictionary
from .resources.resources import available_connectivities, available_stabilizercodes
from .stabilizercode import StabilizerCode
from .connectivity import Connectivity
from .circuit import Circuit
from .symplectic_rep.clifford_gate import Clifford



from typing import Union, Optional, Iterable, Dict, Any
from datetime import datetime
import json
import numpy as np
from numpy.typing import NDArray

try:
    from tqdm import tqdm
except ImportError:
    _has_tqdm = False
finally:
    _has_tqdm = True

from .grb_interface.grb_enviroment import Enviroment
from .grb_interface.grb_gates import *
from .grb_interface.grb_math_interface import *
from .symplectic_rep import *
from .symplectic_rep.integer_symplectic import symplectic_matrix
from .symplectic_rep.helper import LinSolver
from .connectivity import Connectivity
from .stabilizercode import StabilizerCode
from ._utility import MissingOptionalLibraryError


def tailor_logical_gate(
    stab_code: StabilizerCode,
    connectivity: Connectivity,
    logical_gate: Union[Circuit, Clifford, int],
    num_cz_layers: int,
    time_limit: float = -1,
    log_to_console: bool = False,
    log_file: str = "",
    optimize: bool = True,
    gurobi: Dict = {},
    perm: Tuple[bool, bool] = (False, False)
) -> Tuple[Optional[Circuit], str]:
    """
    Finds a circuit implementation for a Clifford gate of a given quantum
    error-correcting code tailored to a specified hardware connectivity.

    Parameters
    ----------
    stab_code: StabilizerCode
        Stabilizer code for which a logical circuit should be tailored.
    connectivity: Connectivity
        Connectivity to tailor the circuit to.
    logical_gate: Union[Circuit, Clifford, int]
        Representation of the logical gate in form of a circuit, Clifford or integer.
    num_cz_layers: int
        Number of controlled-Z gate layers for the ansatz with which the circuit should
        be compiled.
    time_limit: float, optional
        Time in seconds until the program aborts regardless of whether or not a circuit
        implementation has been found. A value of -1 removes the time limit, by default -1.
    log_to_console: bool, optional
        Whether or not Gurobi should log its progress to the console, by default False.
    log_file: str, optional
        File path of the log created by Gurobi. An empty string removes the log file,
        by default "".
    optimize: bool, optional
        Whether to collapse single-qubit Clifford gates after compilation, by default True.
    gurobi: Dict, optional
        Arguments to pass to the Gurobi optimizer, by default {}.
    perm: Tuple[bool, bool]
       If true, a permutation layer is added to the start (index 0) or end (index 1) of the circuit,
        by default [False, False].

    Returns
    -------
    Tuple[Optional[Circuit], str]
        A circuit (if one has been found) and a status message.

    Examples
    --------
        >>> conn = htlg.Connectivity("circular", n=4)
        >>> stab_code = htlg.StabilizerCode("4_2_2")
        >>> log_gate = htlg.Circuit("H 0", 2)
        >>> circ, status = tailor_logical_gate(stab_code, conn, log_gate, 2)

    """
    if not isinstance(stab_code, StabilizerCode):
        raise TypeError("Create qecc object via function 'get_qecc'.")
    stabilizer_matrix = stab_code.get_e_matrix()
    if not isinstance(connectivity, Connectivity):
        raise TypeError("Create connectivity object via function 'get_conn'")
    if not isinstance(logical_gate, Circuit) and not isinstance(logical_gate, int) and not isinstance(logical_gate, Clifford):
        raise TypeError("Wrong type for argument `logical_gate`")

    if isinstance(logical_gate, Circuit):
        logical_gate = logical_gate.to_clifford()
    if isinstance(logical_gate, Clifford):
        add_phases = np.roll(logical_gate.phase,
                             len(stabilizer_matrix[0]) - len(stabilizer_matrix[:, 0]) // 2)
        logical_gate = logical_gate.symplectic_matrix
    else:
        add_phases = None
    if isinstance(logical_gate, int):
        logical_gate = symplectic_matrix(
            logical_gate, len(stabilizer_matrix[:, 0]) // 2)

    if not stab_code.n == connectivity.num_qubits:
        raise ValueError(f"Stabilizer code is defined on '{stab_code.n}' qubits but connectvity on '{connectivity.num_qubits}'")
    if not stab_code.k == len(logical_gate) // 2:
        raise ValueError(f"Stabilizer code has '{stab_code.k}' logical qubits but logical gate is defined on '{len(logical_gate) // 2}'")

    gate_finder = GateFinder(num_cz_layers, connectivity.matrix, stabilizer_matrix, log_to_console,
                             log_file, gurobi, perm)
    if time_limit >= 0:
        gate_finder.set_time_limit(time_limit)
    gate_finder.set_logical_gate(logical_gate)
    gate_finder.set_target_function()
    gate_finder.find_gate()
    if gate_finder.has_solution():
        if add_phases is not None and np.count_nonzero(add_phases) != 0:
            ps = Circuit.from_paulis(
                sum([stabilizer_matrix[:, i] for i in np.nonzero(add_phases)[0]]))
            circuit = ps + gate_finder.get_circuit_implementation()
            if optimize:
                circuit.shallow_optimize()
            return circuit, gate_finder.get_status()
        else:
            circuit = gate_finder.get_circuit_implementation()
            if optimize:
                circuit.shallow_optimize()
            return circuit, gate_finder.get_status()
    else:
        return None, gate_finder.get_status()


def tailor_multiple_logical_gates(
    stab_code: StabilizerCode,
    connectivity: Connectivity,
    logical_gates: Iterable[int],
    num_cz_layers: int,
    output_file: str = "",
    time_limit: float = -1,
    log_file: str = "",
    progress_bar: bool = False,
    save_every: int = 1,
    optimize: bool = True,
    gurobi: Dict = {},
    perm: Tuple[bool, bool] = (False, False)
) -> Dict:
    """
    Finds circuit implementations for multiple Clifford gates of a given quantum
    error-correcting code tailored to a specified hardware connectivity. The
    results can be saved to a file. 

    Parameters
    ----------
    stab_code: StabilizerCode
        Stabilizer code for which a logical circuit should be tailored.
    connectivity: Connectivity
        Connectivity to tailor the circuit to.
    logical_gates: Iterable[int]
        Iterable of integers identifying the Clifford gates that should be constructed.
    num_cz_layers: int
        Number of controlled-Z gate layers for the ansatz with which the circuit should
        be compiled.
    output_file: str, optional
        Filepath where the results dictionary is stored as a .json-file, by default "".
    time_limit: float, optional
        Time in seconds until the program aborts regardless of whether or not a circuit
        implementation has been found. A value of -1 removes the time limit, by default -1.
    log_to_console: bool, optional
        Whether or not Gurobi should log its progress to the console, by default False.
    log_file: str, optional
        File path of the log created by Gurobi. An empty string removes the log file,
        by default "".
    progress_bar: bool, optianl
        Whether or not to show a progress bar. Requires the tqdm package, by default False.
    save_every: int, optional
        If given an output file, continually save gates after constructing this many gates,
        by default 1.
    optimize: bool, optional
        Whether to collapse single-qubit Clifford gates after compilation, by default True.
    gurobi: Dict, optional
        Arguments to pass to the Gurobi optimizer, by default {}.
    perm: Tuple[bool, bool]
       If true, a permutation layer is added to the start (index 0) or end (index 1) of the circuit,
        by default [False, False].

    Returns
    -------
    Dict
        Dictionary containing the constructed circuits and runtime information.
    """
    if not isinstance(stab_code, StabilizerCode):
        raise TypeError("Create qecc object via function 'get_code'!")
    stabilizer_matrix = stab_code.get_e_matrix()
    if not isinstance(connectivity, Connectivity):
        raise TypeError("Create connectivity object via function 'get_conn'!")
    connectivity_matrix = connectivity.matrix

    def iterate(x):
        if progress_bar:
            if not _has_tqdm:
                raise MissingOptionalLibraryError(
                    "progress_bar requires 'qiskit' to be installed")

            return tqdm(x, smoothing=0)  # type: ignore
        return x
    
    if not stab_code.n == connectivity.num_qubits:
        raise ValueError(f"Stabilizer code is defined on '{stab_code.n}' qubits but connectvity on '{connectivity.num_qubits}'")

    gate_finder = GateFinder(num_cz_layers, connectivity_matrix, stabilizer_matrix,
                             False, log_file, gurobi, perm)
    if time_limit >= 0:
        gate_finder.set_time_limit(time_limit)
    gate_finder.set_target_function()
    stor = {
        "Meta": {
            "Connectivity": connectivity,
            "Code": stab_code,
            "Number CZ layers": num_cz_layers,
            "Time limit": time_limit,
            "Started": str(datetime.now())
        },
        "Gates": {}
    }
    for num, i in enumerate(iterate(logical_gates)):
        gate_finder.set_logical_gate(symplectic_matrix(i, gate_finder.k))
        gate_finder.find_gate()
        if gate_finder.has_solution():
            circuit = gate_finder.get_circuit_implementation()
            if optimize:
                circuit.shallow_optimize()
            stor["Gates"][i] = {"Circuit": circuit,
                                "Status": gate_finder.get_status(),
                                "Runtime": gate_finder.get_runtime()}
        else:
            stor["Gates"][i] = {"Circuit": None,
                                "Status": gate_finder.get_status(),
                                "Runtime": gate_finder.get_runtime()}
        if (num + 1) % save_every == 0:
            if output_file != "":
                save_results_dictionary(stor, output_file)
    stor["Meta"]["Ended"] = str(datetime.now())
    if output_file != "":
        save_results_dictionary(stor, output_file)
    return stor

def save_results_dictionary(results: Dict, filepath: str):
    """
    Saves the result dictionary returned by `tailor_multiple_logical_gates`
    to a .json-file.

    Parameters
    ----------
    results: Dict
        Results dictionary.
    filepath: str
        File path to save the .json-file.
    """
    results = deepcopy(results)
    if (c := results.get("Meta", {}).get("Connectivity", "N/A")) != "N/A":
        results["Meta"]["Connectivity"] = c.matrix.tolist()
    if (c := results.get("Meta", {}).get("Code", "N/A")) != "N/A":
        results["Meta"]["Code"] = c.get_e_matrix().tolist()
    for key, val in results.get("Gates", {}).items():
        circ = val.get("Circuit", None) 
        if isinstance(circ, Circuit):
            results["Gates"][key]["Circuit"] = circ.__str__()
    with open(filepath, 'w') as file:
        json.dump(results, file)


def load_results_dictionary(filepath: str):
    """
    Loads the result dictionary saved by `save_results_dictionary`.

    Parameters
    ----------
    filepath: str
        File path to load the results dictionary.
    
    Returns
    ----------
    Dict:
        Dictionary containing the information stored in the .json-file.
    """
    with open(filepath, "r") as file:
        results = json.load(file)
    if (c := results.get("Meta", {}).get("Connectivity", "N/A")) != "N/A":
        results["Meta"]["Connectivity"] = Connectivity(np.array(c, dtype=np.int32)) 
    if (c := results.get("Meta", {}).get("Code", "N/A")) != "N/A":
        results["Meta"]["Code"] = StabilizerCode(np.array(c, dtype=np.int32)) 
    for key, val in results.get("Gates", {}).items():
        circ = val.get("Circuit", None) 
        if isinstance(circ, str) and circ.capitalize() != "NONE":
            results["Gates"][key]["Circuit"] = Circuit(circ)
        else:
            results["Gates"][key]["Circuit"] = None
    for key in list(results.get("Gates", {}).keys()):
        results["Gates"][int(key)] = results["Gates"].pop(key)
    return results


class GateFinder:
    def __init__(
        self,
        num_CZL: int,
        con: NDArray,
        enc: NDArray,
        log_to_console: bool = False,
        log_file: str = "",
        gurobi: Dict[Any, Any] = {},
        perm: Tuple[bool, bool] = (False, False)
    ):
        self.NUM_CZL = num_CZL
        self.CON = con
        self.ENC = Clifford(enc)
        self.n = len(self.CON)
        self.k = self.ENC.m - self.n
        self.active_gate = False

        self.env = Enviroment(log_to_console, log_file, gurobi)
        self.SCLs = [create_SCL(self.n, self.env)] + \
            [create_cons_SCL(self.n, self.env) for _ in range(self.NUM_CZL)]
        self.CZLs = [create_CZL(self.CON, self.env)
                     for _ in range(self.NUM_CZL)]
        self.LOGICAL, self.LOG_IDS = self.env.create_predef_bin_matrix(
            2*self.k, 2*self.k)
        self.FREEDOM = create_reduced_freedom_matrix(self.n, self.k, self.env)
        self.Perms: List[Union[None, Clifford, ExprMatrix]] = [None, None]
        self.ANSATZ = self.SCLs[0]
        if perm[1]:
            self.Perms[1] = create_Perm(self.n, self.env)
            self.ANSATZ = self.Perms[1] @ self.ANSATZ
        for scl, czl in zip(self.SCLs[1:], self.CZLs):
            self.ANSATZ = self.ANSATZ @ czl @ scl
        if perm[0]:
            self.Perms[0] = create_Perm(self.n, self.env)
            self.ANSATZ = self.ANSATZ @ self.Perms[0]
        enc_expr = ExprMatrix.create_from_array(self.ENC.symplectic_matrix)
        self.env.add_equality_constraint_mat(
            enc_expr @ (self.LOGICAL.create_expanded_dims(self.n +
                        self.k, self.n + self.k) + self.FREEDOM),
            self.ANSATZ @ enc_expr,
            True
        )
        self.lin_solv = LinSolver(enc.T)

    def set_time_limit(self, time_limit: float):
        self.env.set_time_limit(time_limit)

    def set_logical_gate(self, logical_gate: NDArray):
        self.env.set_many_predef_var(logical_gate % 2, self.LOG_IDS)
        self.active_gate = True

    def set_target_function(self):
        e = Expression.create_const(0)
        for czl in self.CZLs:
            for i in range(self.n):
                for j in range(i+1, self.n):
                    e = e + czl[self.n+i, j]
        self.env.set_target_function(e)

    def has_solution(self) -> bool:
        return self.env.has_solution()

    def get_status(self) -> str:
        return self.env.get_status()

    def get_runtime(self) -> float:
        return self.env.get_runtime()

    def get_work(self) -> float:
        return self.env.get_work()

    def find_gate(self):
        assert (self.active_gate)
        self.env.solve()

    def get_circuit_implementation(self) -> Circuit:
        cliffs: List[Clifford] = [None] * (2*self.NUM_CZL + 1)
        circs: List[Circuit] = [None] * (2*self.NUM_CZL + 1)
        cliffs[::2] = [Clifford(
            self.env.evaluate_matrix(i)) for i in self.SCLs]
        cliffs[1::2] = [Clifford(
            self.env.evaluate_matrix(i)) for i in self.CZLs]
        tot_cliff: Clifford = cliffs[0]
        circs[::2] = [Circuit.from_SCL(i) for i in cliffs[::2][::-1]]
        circs[1::2] = [Circuit.from_CZL(i) for i in cliffs[1::2][::-1]]

        for c in cliffs[1:]:
            tot_cliff = tot_cliff @ c
        if self.Perms[0] != None:
            self.Perms[0] = Clifford(
                self.env.evaluate_matrix(self.Perms[0]))
            tot_cliff = tot_cliff @ self.Perms[0]
            circs.insert(0, Circuit.from_permutation(self.Perms[0]))
        if self.Perms[1] != None:
            self.Perms[1] = Clifford(
                self.env.evaluate_matrix(self.Perms[1]))
            tot_cliff = self.Perms[1] @ tot_cliff
            circs.append(Circuit.from_permutation(self.Perms[1]))

        paulis = Circuit.from_paulis(
            self.lin_solv.get_solution((tot_cliff@self.ENC).phase), invert=True)
        return sum(circs, start=paulis)



import unittest

import numpy as np

from htlogicalgates.connectivity import *


class TestConnenctivity(unittest.TestCase):
    def test_constructor(self):
        arr = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]], dtype=np.int32)
        c = Connectivity("linear", 3)
        np.testing.assert_equal(c.matrix, arr)
        self.assertEqual(c.num_qubits, 3)
        np.testing.assert_equal(Connectivity(arr).matrix, arr)



# canonical ordering of symplectic group elements
# from ”How to efficiently select an arbitrary clifford group element”
# by Robert Koenig and John A. Smolin

from numpy import *
from random import randint, shuffle


def directsum(m1, m2):
    n1 = len(m1[0])
    n2 = len(m2[0])
    output = zeros((n1+n2, n1+n2), dtype=int32)
    for i in range(0, n1):
        for j in range(0, n1):
            output[i, j] = m1[i, j]
    for i in range(0, n2):
        for j in range(0, n2):
            output[i+n1, j+n1] = m2[i, j]
    return output


def inner(v, w):
    t = 0
    for i in range(0, size(v) >> 1):
        t += v[2*i]*w[2*i+1]
        t += w[2*i]*v[2*i+1]
    return t % 2


def transvection(k, v):
    return (v+inner(k, v)*k) % 2


def int2bits(i, n):
    # i = int(i)
    output = zeros(n, dtype=int32)
    for j in range(0, n):
        output[j] = i & 1
        i >>= 1
    return output


def findtransvection(x, y):
    output = zeros((2, size(x)), dtype=int32)
    if array_equal(x, y):
        return output
    if inner(x, y) == 1:
        output[0] = (x+y) % 2
        return output
    z = zeros(size(x))
    for i in range(0, size(x) >> 1):
        ii = 2*i
        if ((x[ii]+x[ii + 1]) != 0) and ((y[ii]+y[ii + 1]) != 0):
            z[ii] = (x[ii]+y[ii]) % 2
            z[ii + 1] = (x[ii + 1]+y[ii + 1]) % 2
            if (z[ii] + z[ii + 1]) == 0:  # they were the same so they added t o 00
                z[ii + 1] = 1
                if x[ii] != x[ii + 1]:
                    z[ii] = 1
            output[0] = (x+z) % 2
            output[1] = (y+z) % 2
            return output
    for i in range(0, size(x) >> 1):
        ii = 2 * i
        if ((x[ii]+x[ii + 1]) != 0) and ((y[ii]+y[ii + 1]) == 0):  # found the p ai r
            if x[ii] == x[ii + 1]:
                z[ii + 1] = 1
            else:
                z[ii + 1] = x[ii]
                z[ii] = x[ii + 1]
            break
    for i in range(0, size(x) >> 1):
        ii = 2 * i
        if ((x[ii]+x[ii + 1]) == 0) and ((y[ii]+y[ii + 1]) != 0):  # found the p ai r
            if y[ii] == y[ii + 1]:
                z[ii + 1] = 1
            else:
                z[ii + 1] = y[ii]
                z[ii] = y[ii + 1]
            break
    output[0] = (x+z) % 2
    output[1] = (y+z) % 2
    return output


def symplectic(i, n):
    nn = 2*n
    s = ((1 << nn)-1)
    k = (i % s)+1
    i //= s  # i/=s
    # i = int(round(i))
    f1 = int2bits(k, nn)
    e1 = zeros(nn, dtype=int32)
    e1[0] = 1
    T = findtransvection(e1, f1)
    bits = int2bits(i % (1 << (nn-1)), nn-1)
    eprime = copy(e1)
    for j in range(2, nn):
        eprime[j] = bits[j-1]
    h0 = transvection(T[0], eprime)
    h0 = transvection(T[1], h0)
    if bits[0] == 1:
        f1 *= 0
    id2 = zeros((2, 2), dtype=int32)
    id2[0, 0] = 1
    id2[1, 1] = 1
    if n != 1:
        g = directsum(id2, symplectic(i >> (nn-1), n-1))
    else:
        g = id2
    for j in range(0, nn):
        g[j] = transvection(T[0], g[j])
        g[j] = transvection(T[1], g[j])
        g[j] = transvection(h0, g[j])
        g[j] = transvection(f1, g[j])
    return g


def bits2int(b, nn):
    output = 0
    tmp = 1
    for j in range(nn):
        if b[j] == 1:
            output = output + tmp
        tmp = tmp*2
    return output


def numberofcosets(n):
    x = power(2, 2*n-1)*(power(2, 2*n)-1)
    return x


def numberofsymplectic(n):
    x = 1
    for j in range(1, n+1):
        x = x*numberofcosets(j)
    return x


def symplecticinverse(n, gn):
    nn = 2*n
    v = gn[0]
    w = gn[1]

    e1 = zeros(nn, dtype=int32)
    e1[0] = 1
    T = findtransvection(v, e1)

    tw = copy(w)
    tw = transvection(T[0], tw)
    tw = transvection(T[1], tw)
    b = tw[0]
    h0 = zeros(nn, dtype=int32)
    h0[0] = 1
    h0[1] = 0
    for j in range(2, nn):
        h0[j] = tw[j]

    bb = zeros(nn-1, dtype=int32)
    bb[0] = b
    for j in range(2, nn):
        bb[j-1] = tw[j]
    zv = bits2int(v, nn) - 1
    zw = bits2int(bb, nn-1)
    cvw = zw*(power(2, 2*n) - 1)+zv

    if n == 1:
        return cvw

    gprime = copy(gn)
    if b == 0:
        for j in range(nn):
            gprime[j] = transvection(T[1], transvection(T[0], gn[j]))
            gprime[j] = transvection(h0, gprime[j])
            gprime[j] = transvection(e1, gprime[j])
    else:
        for j in range(nn):
            gprime[j] = transvection(T[1], transvection(T[0], gn[j]))
            gprime[j] = transvection(h0, gprime[j])

    gnew = gprime[2:nn, 2:nn]
    gnidx = symplecticinverse(n-1, gnew)*numberofcosets(n)+cvw
    return gnidx


###################
###################
###################


_traf_mat = {}


def _get_traf_mat(n: int):
    t = zeros((2*n, 2*n), dtype=int32)
    for i in range(n):
        t[i, 2*i] = t[-i-1, -2*i-1] = 1
    return t


def symplectic_matrix(i: int, n: int):
    if i == -1:
        pr = 1
        for j in range(n):
            pr *= 4**(j+1) - 1
        i = randint(0, 2**(n**2)*pr - 1)
    global _traf_mat
    if n not in _traf_mat.keys():
        _traf_mat[n] = _get_traf_mat(n)
    return _traf_mat[n]@symplectic(i, n)@_traf_mat[n].transpose()


def symplectic_matrix_inverse(gn: ndarray, n: int):
    global _traf_mat
    if n not in _traf_mat.keys():
        _traf_mat[n] = _get_traf_mat(n)
    return symplecticinverse(n, _traf_mat[n].transpose()@gn@_traf_mat[n])


def is_symplectic(arr: ndarray):
    n = arr.shape[0]//2
    O = zeros((2*n, 2*n), dtype=int32)
    O[:n, n:2*n] = O[n:2*n, :n] = eye(n, dtype=int32)
    return all((arr.transpose()@O@arr) % 2 == O)


def expand_symplectic(arr: ndarray, b: int, a: int):
    c = arr.shape[0]//2
    new_d = c + b + a
    new_arr = eye(2*new_d, dtype=int32)
    new_arr[b: c+b, b: c+b] = arr[:c, :c]
    new_arr[b+new_d: c+b+new_d, b: c+b] = arr[c:2*c, :c]
    new_arr[b: c+b, b+new_d: c+b+new_d] = arr[:c, c:2*c]
    new_arr[b+new_d: c+b+new_d, b+new_d: c+b+new_d] = arr[c:2*c, c:2*c]
    return new_arr



from __future__ import annotations
from enum import Enum
from typing import Tuple, List, Union, overload
import numpy as np
from numpy.typing import NDArray

try:
    from qiskit import QuantumCircuit as qiskitQuantumCircuit
except ImportError:
    _has_qiskit = False
finally:
    _has_qiskit = True

from .symplectic_rep.clifford_gate import Clifford
from .symplectic_rep.helper import matrix_rank, get_AMG_decomposition
from ._utility import _argument_assignment, MissingOptionalLibraryError

# Identical to stim circuit language


class Operation(Enum):
    CZ = "CZ"
    CX = "CX"
    I = "I"
    S = "S"
    SDG = "SDG"
    H = "H"
    C_ZYX = "C_ZYX"  # S H, X<-Y<-Z<-X
    C_XYZ = "C_XYZ"  # H S_DAG, X->Y->Z->X
    SXDG = "SQRT_X_DAG"  # SQRT_X_DAG = H S_DAG H
    SWAP = "SWAP"
    BARRIER = ""
    X = "X"
    Y = "Y"
    Z = "Z"


Gate = Tuple[Operation, List[int]]


def contract_single_qubit_clifford(ops: List[Operation]) -> List[Operation]:
    if len(ops) == 0:
        return [Operation.I]
    c = gate_to_clifford(ops[0], [0], 1)
    for op in ops[1:]:
        c = gate_to_clifford(op, [0], 1) @ c
    circ = Circuit.from_SCL(c)
    return [i[0] for i in circ._gates]


def gate_to_clifford(op: Operation, qubits: List[int], num_qubits: int):
    n = num_qubits
    if op == Operation.CZ:
        assert (len(qubits) == 2)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0]+n, qubits[1]] = m[qubits[1]+n, qubits[0]] = 1
        return Clifford(m)
    if op == Operation.CX:
        assert (len(qubits) == 2)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[1], qubits[0]] = m[qubits[0]+n, qubits[1]+n] = 1
        return Clifford(m)
    if op == Operation.SWAP:
        assert (len(qubits) == 2)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0], qubits[0]] = m[qubits[0]+n, qubits[0]+n] = 0
        m[qubits[1], qubits[1]] = m[qubits[1]+n, qubits[1]+n] = 0
        m[qubits[0], qubits[1]] = m[qubits[0]+n, qubits[1]+n] = 1
        m[qubits[1], qubits[0]] = m[qubits[1]+n, qubits[0]+n] = 1
        return Clifford(m)
    if op == Operation.I or op == Operation.BARRIER:
        m = np.identity(2*n, dtype=np.int32)
        return Clifford(m)
    if op == Operation.S:
        assert (len(qubits) == 1)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0]+n, qubits[0]] = 1
        return Clifford(m)
    if op == Operation.H:
        assert (len(qubits) == 1)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0]+n, qubits[0]] = m[qubits[0], qubits[0]+n] = 1
        m[qubits[0], qubits[0]] = m[qubits[0]+n, qubits[0]+n] = 0
        return Clifford(m)
    if op == Operation.C_ZYX:
        assert (len(qubits) == 1)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0]+n, qubits[0]] = m[qubits[0], qubits[0]+n] = 1
        m[qubits[0], qubits[0]] = 0
        return Clifford(m)
    if op == Operation.C_XYZ:
        assert (len(qubits) == 1)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0]+n, qubits[0]] = m[qubits[0], qubits[0]+n] = 1
        m[qubits[0]+n, qubits[0]+n] = 0
        return Clifford(m)
    if op == Operation.SXDG:
        assert (len(qubits) == 1)
        m = np.identity(2*n, dtype=np.int32)
        m[qubits[0], qubits[0]+n] = 1
        return Clifford(m)
    if op == Operation.X:
        assert (len(qubits) == 1)
        p = np.zeros((2*n,), dtype=np.int32)
        p[qubits[0]+n] = 1
        return Clifford(np.identity(2*n, dtype=np.int32), p)
    if op == Operation.Y:
        assert (len(qubits) == 1)
        p = np.zeros((2*n,), dtype=np.int32)
        p[qubits[0]+n] = p[qubits[0]] = 1
        return Clifford(np.identity(2*n, dtype=np.int32), p)
    if op == Operation.Z:
        assert (len(qubits) == 1)
        p = np.zeros((2*n,), dtype=np.int32)
        p[qubits[0]] = 1
        return Clifford(np.identity(2*n, dtype=np.int32), p)
    raise ValueError(f"Operation '{op.value}' not known")


class Circuit:
    @overload
    def __init__(self, num_qubits: int):
        """
        Constructs an empty circuit.

        Parameters
        ----------
        num_qubits: int
            Number of qubits the circuit acts on.
        """
        pass

    @overload
    def __init__(self, init_string: str):
        """
        Constructs a circuit from a string.

        Parameters
        ----------
        init_string: str
            String containing the circuit.

        Examples
        ----------
            >>> Circuit("H 0\\nS 1\\nCZ 0 1")

        """
        pass

    @overload
    def __init__(self, init_string: str, num_qubits: int):
        """
        Constructs a circuit from a string defined on a number of qubits. The number of
        qubits must be large enough to contain all gates from the string.

        Parameters
        ----------
        init_string: str
            String containing the circuit.
        num_qubits: int
            Number of qubits the circuit acts on.

        Examples
        ----------
            >>> Circuit("H 0\\nS 1\\nCZ 0 1", 4)
        """
        pass

    @overload
    def __init__(self, cliff: Clifford):
        """
        Construct a circuit from a Clifford gate using the algorithm
        from <https://arxiv.org/pdf/quant-ph/0406196>.

        Parameters
        ----------
        cliff: Clifford
            Clifford element to create a circuit from.
        """
        pass

    def __init__(self, *args, **kwargs):
        options = [{"num_qubits": int},
                   {"init_string": str},
                   {"init_string": str, "num_qubits": int},
                   {"cliff": Clifford}]
        i, a = _argument_assignment(
            options, "Circuit()", *args, **kwargs)
        if i == 0:
            self._num_qubits = a["num_qubits"]
            self._gates: List[Gate] = []
        elif i == 1 or i == 2:
            self._gates: List[Gate] = []
            m = 0
            try:
                for j, l in enumerate(a["init_string"].splitlines()):
                    l = l.split("#", 1)[0]
                    parts = l.strip().split()
                    if len(parts) == 0:
                        continue
                    elif len(parts) == 1:
                        for operation in Operation:
                            if operation.value in parts[0].upper():
                                self._gates.append(
                                    (operation, [int(parts[0].upper().replace(operation.value, ""))]))
                                m = max(max(self._gates[-1][1])+1, m)
                                break
                    else:
                        self._gates.append((Operation(parts[0].upper()), [
                            int(i) for i in parts[1:]]))
                        m = max(max(self._gates[-1][1])+1, m)
            except ValueError:
                raise ValueError(
                    f"Invalid instruction in line {str(j)}: '{l}'")
            if i == 2:
                self._num_qubits = a["num_qubits"]
                if m > a["num_qubits"]:
                    raise ValueError(
                        f"Circuit is defined on at least '{m}' qubits but only '{a['num_qubits']}' were given")
            else:
                self._num_qubits = m
        elif i == 3:
            # from <https://arxiv.org/pdf/quant-ph/0406196> Theorem 8
            cliff: Clifford = a["cliff"]
            self._num_qubits = cliff.num_qubits
            self._gates: List[Gate] = []
            n = cliff.num_qubits

            # Step 1
            def get_rank(x): return matrix_rank(
                x.symplectic_matrix[0:n, n:2*n])
            prev_rank = get_rank(cliff)
            for i in range(n):
                h = gate_to_clifford(Operation.H, [i], n)
                if (new_rank := get_rank(h@cliff)) > prev_rank:
                    cliff = h@cliff
                    self.insert(0, (Operation.H, [i]))
                    prev_rank = new_rank

            # Step 2
            for i in range(n):
                if cliff.symplectic_matrix[0:n, n:2*n][i, i] % 2 == 0:
                    for j in range(i+1, n):
                        if cliff.symplectic_matrix[0:n, n:2*n][j, i] % 2 == 1:
                            cliff = gate_to_clifford(
                                Operation.CX, [j, i], n)@cliff
                            self.insert(0, (Operation.CX, [j, i]))
                            break
                for j in range(i+1, n):
                    if cliff.symplectic_matrix[0:n, n:2*n][j, i] % 2 == 1:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))
            for i in reversed(range(n)):
                for j in range(0, i):
                    if cliff.symplectic_matrix[0:n, n:2*n][j, i] % 2 == 1:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))

            # Step 3
            M, G = get_AMG_decomposition(cliff.symplectic_matrix[n:2*n, n:2*n])
            for i in range(n):
                if G[i, i] % 2 == 1:
                    cliff = gate_to_clifford(Operation.S, [i], n)@cliff
                    self.insert(0, (Operation.S, [i]))

            # Step 4
            for i in range(n):
                for j in range(0, i):
                    if M[i, j] % 2 == 1:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))

            # Step 5
            for i in range(n):
                cliff = gate_to_clifford(Operation.S, [i], n)@cliff
                self.insert(0, (Operation.S, [i]))

            # Step 6
            for i in reversed(range(n)):
                for j in range(0, i):
                    if cliff.symplectic_matrix[:n, n:2*n][j, i]:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))

            # Step 7
            for i in range(n):
                cliff = gate_to_clifford(Operation.H, [i], n)@cliff
                self.insert(0, (Operation.H, [i]))

            # Step 8
            M, G = get_AMG_decomposition(cliff.symplectic_matrix[n:2*n, :n])
            for i in range(n):
                if G[i, i] % 2 == 1:
                    cliff = gate_to_clifford(Operation.S, [i], n)@cliff
                    self.insert(0, (Operation.S, [i]))

            # Step 9
            for i in range(n):
                for j in range(0, i):
                    if M[i, j] % 2 == 1:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))

            # Step 10
            for i in range(n):
                cliff = gate_to_clifford(Operation.S, [i], n)@cliff
                self.insert(0, (Operation.S, [i]))

            # Step 11
            for i in reversed(range(n)):
                for j in range(0, i):
                    if M[i, j] % 2 == 1:
                        cliff = gate_to_clifford(Operation.CX, [i, j], n)@cliff
                        self.insert(0, (Operation.CX, [i, j]))

            # Fix Pauli frame
            circuit_clifford = self.to_clifford()
            target_clifford = a["cliff"]
            phases = (circuit_clifford.phase + target_clifford.phase) % 2
            for i in range(n):
                if phases[i] == 1:
                    self.insert(0, (Operation.Z, [i]))
                if phases[i + n] % 2 == 1:
                    self.insert(0, (Operation.X, [i]))

    def h(self, qubit: int):
        """
        Adds a Hadamard gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.H, [qubit]))

    def s(self, qubit: int):
        """
        Adds a phase gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.S, [qubit]))

    def sdg(self, qubit: int):
        """
        Adds the adjoint of the phase gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.SDG, [qubit]))

    def sxdg(self, qubit: int):
        """
        Adds a sqrt(X)=(h sdg h) gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.SXDG, [qubit]))

    def c_xyz(self, qubit: int):
        """
        Adds a c_xyz=(h sdg) gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.C_XYZ, [qubit]))

    def c_zyx(self, qubit: int):
        """
        Adds a c_zyx=(s h) gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.C_ZYX, [qubit]))

    def x(self, qubit: int):
        """
        Adds a Pauli-X gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.X, [qubit]))

    def y(self, qubit: int):
        """
        Adds a Pauli-Y gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.Y, [qubit]))

    def z(self, qubit: int):
        """
        Adds a Pauli-Z gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.Z, [qubit]))

    def id(self, qubit: int):
        """
        Adds an identity gate to the end of circuit.

        Parameters
        ----------
        qubit: int
            Qubit the gate acts on.
        """
        self.append((Operation.I, [qubit]))

    def cx(self, control: int, target: int):
        """
        Adds a controlled-X gate to the end of circuit.

        Parameters
        ----------
        control: int
            Control qubit.
        target: int
            Target qubit.
        """
        self.append((Operation.CX, [control, target]))

    def cz(self, qubit1: int, qubit2):
        """
        Adds a controlled-Z gate to the end of circuit.

        Parameters
        ----------
        qubit1: int
            First qubit the gate acts on.
        qubit2: int
            Second qubit the gate acts on.
        """
        self.append((Operation.CZ, [qubit1, qubit2]))

    def swap(self, qubit1: int, qubit2):
        """
        Adds a swapping operation to the end of circuit.

        Parameters
        ----------
        qubit1: int
            First qubit the gate acts on.
        qubit2: int
            Second qubit the gate acts on.
        """
        self.append((Operation.SWAP, [qubit1, qubit2]))

    def __add__(self, other: Circuit) -> Circuit:
        assert (self.num_qubits == other.num_qubits)
        circuit = Circuit(self.num_qubits)
        circuit.append(self._gates + other._gates)
        return circuit

    @property
    def num_qubits(self) -> int:
        """
        Returns the number of qubits the Clifford is defined on.

        Returns
        ----------
        int
            Number of qubits.
        """
        return self._num_qubits

    def shallow_optimize(self):
        """
        Performs a small optimization of single-qubit gates of the circuit.
        This operation is in-place. 
        """
        for i in range(self.num_qubits):
            targets = [[]]
            ops = [[]]
            for j, gate in enumerate(self._gates):
                if gate[1] == [i]:
                    targets[-1].append(j)
                    ops[-1].append(gate[0])
                elif len(gate) == 2 and i in gate[1]:
                    targets.append([])
                    ops.append([])
            for qubits, os in zip(reversed(targets), reversed(ops)):
                if len(qubits) == 0:
                    continue
                o = contract_single_qubit_clifford(os)
                for j in reversed(qubits):
                    self._gates.pop(j)
                if len(o) == 1 and o[0] == Operation.I:
                    continue
                for el in reversed(o):
                    self._gates.insert(qubits[0], (el, [i]))
        # Collect Paulis

    def permuted(self, map: dict, num_qubits: Union[int, None] = None) -> Circuit:
        if num_qubits is None:
            num_qubits = self.num_qubits
        circuit = Circuit(num_qubits)
        for op, qubits in self._gates:
            circuit.append((op, [map.get(i, i) for i in qubits]))
        return circuit

    def two_qubit_gate_count(self) -> int:
        """
        Returns the number of two-qubit gates in the circuit.

        Returns
        ----------
        int
            Number of two-qubit gates.
        """
        i = 0
        for op, qubits in self._gates:
            if len(qubits) == 2:
                i += 1
        return i

    def gate_count(self) -> int:
        """
        Returns the number of gates in the circuit.

        Returns
        ----------
        int
            Number of gates.
        """
        return len(self._gates)

    def append(self, gate: Union[Gate, List[Gate]]):
        """
        Adds one or multiple gates to the circuit.
        In order to compose two circuits, use `Circuit("H 0") + Circuit("S 0")`.

        Parameters
        ----------
        gate: Union[Gate, List[Gate]]
            Gate or list of gates to add to the end of the circuit.
        """
        if isinstance(gate, list):
            self._gates += gate
        else:
            self._gates.append(gate)

    def insert(self, index: int, gate: Gate):
        """
        Adds a gate to a specified position in the circuit.
        In order to compose two circuits, use `Circuit("H 0") + Circuit("S 0")`.

        Parameters
        ----------
        index: int
            Position of the gate.
        gate: Gate
            Gate to add to the circuit.
        """
        self._gates.insert(index, gate)

    def __str__(self) -> str:
        s = ""
        for op, qubits in self._gates:
            s += op.value
            for t in qubits:
                s += f" {str(t)}"
            s += "\n"
        return s

    def to_qiskit(self):
        """
        Transforms the circuit into a qiskit circuit.

        Raises

        Returns
        ----------
        qiskit.QuantumCircuit
            Circuit as QuantumCircuit from qiskit.

        Raises
        ----------
        MissingOptionalLibraryError
            If qiskit is not installed
        """
        if not _has_qiskit:
            raise MissingOptionalLibraryError(
                "to_qiskit() requires 'qiskit' to be installed")
        QC = qiskitQuantumCircuit
        gates = {
            Operation.X: QC.x, Operation.Y: QC.y, Operation.Z: QC.z, Operation.H: QC.h,
            Operation.SDG: QC.sdg, Operation.S: QC.s, Operation.CX: QC.cx,
            Operation.CZ: QC.cz, Operation.SXDG: QC.sxdg,
            Operation.SWAP: QC.swap, Operation.I: QC.id
        }

        circuit = QC(self.num_qubits)

        for op, qubits in self._gates:
            if op in gates:
                gates[op](circuit, *qubits)
            elif op == Operation.C_XYZ:
                circuit.sdg(*qubits)
                circuit.h(*qubits)
            elif op == Operation.C_ZYX:
                circuit.h(*qubits)
                circuit.s(*qubits)
            elif op == Operation.BARRIER:
                circuit.barrier()
            else:
                assert False, f"Unknown op {op}"
        return circuit

    def to_clifford(self) -> Clifford:
        """
        Returns the circuit as a Clifford object.

        Returns
        ----------
        Clifford
            Clifford object of the circuit.
        """
        c = Clifford(np.identity(
            2*self.num_qubits, dtype=np.int32))
        for gate, ts in reversed(self._gates):
            c = c@gate_to_clifford(gate, ts, self.num_qubits)
        return c

    @staticmethod
    def from_CZL(clifford: Clifford) -> Circuit:
        circ = Circuit(clifford.num_qubits)
        for i in range(clifford.num_qubits):
            for j in range(i+1, clifford.num_qubits):
                if clifford.symplectic_matrix[i+clifford.num_qubits, j] == 1:
                    circ.append((Operation.CZ, [i, j]))
        return circ

    @staticmethod
    def from_SCL(clifford: Clifford, only_paulis: bool = False) -> Circuit:
        circ = Circuit(clifford.num_qubits)
        for i in range(clifford.num_qubits):
            if clifford.phase[i] == 1:
                circ.append((Operation.Z, [i]))
            if clifford.phase[i+clifford.num_qubits] == 1:
                circ.append((Operation.X, [i]))
        if only_paulis:
            return circ
        for i in range(clifford.num_qubits):
            m = (clifford.symplectic_matrix[i, i], clifford.symplectic_matrix[i, i+clifford.num_qubits],
                 clifford.symplectic_matrix[i+clifford.num_qubits, i], clifford.symplectic_matrix[i+clifford.num_qubits, i+clifford.num_qubits])
            if m == (1, 0, 0, 1):
                pass
                # circ.append((Operation.I, [i]))
            elif m == (1, 0, 1, 1):
                circ.append((Operation.S, [i]))
            elif m == (1, 1, 0, 1):
                circ.append((Operation.SXDG, [i]))
            elif m == (0, 1, 1, 0):
                circ.append((Operation.H, [i]))
            elif m == (1, 1, 1, 0):
                circ.append((Operation.C_XYZ, [i]))
            elif m == (0, 1, 1, 1):
                circ.append((Operation.C_ZYX, [i]))
            else:
                raise ValueError(
                    f"Gate with signature {str(m)} at qubit {str(i)} is not Clifford")
        return circ

    @staticmethod
    def from_permutation(clifford: Clifford) -> Circuit:
        circ = Circuit(clifford.num_qubits)
        for i in range(clifford.num_qubits):
            if clifford.symplectic_matrix[i, i] == 0:
                for j in range(clifford.num_qubits):
                    if clifford.symplectic_matrix[j, i] == 1:
                        circ.append((Operation.SWAP, [i, j]))
                        clifford = clifford @ gate_to_clifford(
                            Operation.SWAP, [i, j], clifford.num_qubits
                        )
                        break
        return circ

    @staticmethod
    def from_paulis(paulis: NDArray, invert: bool = False) -> Circuit:
        circ = Circuit(len(paulis) // 2)
        o1, o2 = (0, len(paulis) // 2) if invert else (len(paulis) // 2, 0)
        for i in range(len(paulis) // 2):
            if paulis[i + o1] == 1:
                circ.append((Operation.Z, [i]))
            if paulis[i + o2] == 1:
                circ.append((Operation.X, [i]))
        return circ



import unittest
import os
from tempfile import TemporaryDirectory

from htlogicalgates import *


class TestTailorLogicalGate(unittest.TestCase):
    def test_tailor_logical_gate(self):
        conn = Connectivity("circular", num_qubits=4)
        qecc = StabilizerCode("4_2_2")
        log_gate = Circuit("H 0", 2)
        circ, status = tailor_logical_gate(qecc, conn, log_gate, 2)
        self.assertEqual(status, "Optimal")
        self.assertEqual(circ.two_qubit_gate_count(), 4)

    def test_tailor_logical_gate_phase(self):
        conn = Connectivity("all", num_qubits=3)
        qecc = StabilizerCode("trivial", num_qubits=3)
        log_gate = Circuit("H 2\nX 0\nCZ 0 1\nCZ 0 2\nY 1\nH 0\nS 1\nCX 1 0\nY 1\nH 1\nX 2", 3).to_clifford()
        circ, status = tailor_logical_gate(qecc, conn, log_gate, 2)
        self.assertEqual(status, "Optimal")
        self.assertTrue(circ.to_clifford() == log_gate)

    def test_tailor_logical_gate_checks(self):
        conn = Connectivity("circular", num_qubits=5)
        qecc = StabilizerCode("4_2_2")
        log_gate = Circuit("H 0", 2)
        self.assertRaises(ValueError, lambda: tailor_logical_gate(
            qecc, conn, log_gate, 2))
        self.assertRaises(ValueError, lambda: tailor_multiple_logical_gates(
            qecc, conn, [0, 1, 121], 2))

        conn = Connectivity("circular", num_qubits=4)
        qecc = StabilizerCode("4_2_2")
        log_gate = Circuit("H 0", 3)
        self.assertRaises(ValueError, lambda: tailor_logical_gate(
            qecc, conn, log_gate, 2))

    def test_tailor_multiple_logical_gates(self):
        conn = Connectivity("circular", num_qubits=4)
        qecc = StabilizerCode("4_2_2")
        res = tailor_multiple_logical_gates(qecc, conn, [0, 1, 121], 2)
        self.assertEqual(len(res["Gates"]), 3)
        self.assertEqual(res["Gates"][0]["Circuit"].two_qubit_gate_count(), 0)
        self.assertEqual(res["Gates"][1]["Circuit"].two_qubit_gate_count(), 4)
        self.assertEqual(res["Gates"][121]
                         ["Circuit"].two_qubit_gate_count(), 3)


class TestSaveLoadResults(unittest.TestCase):
    def test_save_load_results_dictionary(self):
        conn = Connectivity("circular", num_qubits=4)
        qecc = StabilizerCode("4_2_2")
        res = tailor_multiple_logical_gates(qecc, conn, [0, 1, 121], 2)
        with TemporaryDirectory() as tmp_dir:
            filepath = os.path.join(tmp_dir, "results.json")
            save_results_dictionary(res, filepath)
            res_loaded = load_results_dictionary(filepath)
        self.assertEqual(res_loaded["Gates"][0]["Circuit"].two_qubit_gate_count(),
                         res["Gates"][0]["Circuit"].two_qubit_gate_count())
        self.assertEqual(res_loaded["Gates"][1]["Circuit"].two_qubit_gate_count(),
                         res["Gates"][1]["Circuit"].two_qubit_gate_count())
        self.assertEqual(res_loaded["Gates"][121]["Circuit"].two_qubit_gate_count(),
                         res["Gates"][121]["Circuit"].two_qubit_gate_count())
        self.assertEqual(res_loaded["Meta"]["Number CZ layers"], 2)
        self.assertIsInstance(res_loaded["Meta"]["Connectivity"], Connectivity)
        self.assertIsInstance(res_loaded["Meta"]["Code"], StabilizerCode)
        self.assertEqual(res_loaded["Meta"]["Connectivity"].num_qubits, 4)
        self.assertEqual(res_loaded["Meta"]["Code"].k, 2)

    def test_save_load_none(self):
        conn = Connectivity("circular", num_qubits=4)
        qecc = StabilizerCode("4_2_2")
        res = tailor_multiple_logical_gates(qecc, conn, [0, 1, 121], 2, time_limit=0.001)
        with TemporaryDirectory() as tmp_dir:
            filepath = os.path.join(tmp_dir, "results.json")
            save_results_dictionary(res, filepath)
            res_loaded = load_results_dictionary(filepath)
        self.assertEqual(res_loaded["Gates"][0]["Circuit"],
                         res["Gates"][0]["Circuit"])
        self.assertEqual(res_loaded["Gates"][1]["Circuit"],
                         res["Gates"][1]["Circuit"])
        self.assertEqual(res_loaded["Gates"][121]["Circuit"],
                         res["Gates"][121]["Circuit"])
        self.assertEqual(res_loaded["Meta"]["Number CZ layers"], 2)
        self.assertIsInstance(res_loaded["Meta"]["Connectivity"], Connectivity)
        self.assertIsInstance(res_loaded["Meta"]["Code"], StabilizerCode)
        self.assertEqual(res_loaded["Meta"]["Connectivity"].num_qubits, 4)
        self.assertAlmostEqual(res_loaded["Meta"]["Time limit"], 0.001)
        self.assertEqual(res_loaded["Meta"]["Code"].k, 2)



ENCODING_FILE = "encodings.json"
CON_FILE = "connectivities.json"
ENCODING_KEY = "enc_t"
CON_KEY = "con"
DESCR_KEY = "desc"


class MissingOptionalLibraryError(Exception):
    pass


def _argument_assignment(options, name, *args, **kwargs):
    class Empty:
        ...

    def try_option(option):
        if len(args) + len(kwargs) != len(option):
            return None
        e = Empty()
        res = dict.fromkeys(option.keys(), e)
        for key, val in zip(res.keys(), args):
            if not isinstance(val, option[key]):
                return None
            else:
                res[key] = val
        for key, val in kwargs.items():
            if not key in res.keys():
                return None
            if res[key] != e:
                raise TypeError(
                    f"{name} got multiple values for argument '{key}'")
            if not isinstance(val, option[key]):
                raise TypeError(f"{name} argument '{key}' is of type '{str(type(val))}', must be of type '{str(option[key])}'")
            res[key] = val
        if not any([i is e for i in res.values()]):
            return res
        return None

    for i, option in enumerate(options):
        result = try_option(option)
        if result != None:
            return i, result
    raise TypeError(f"{name} got invalid arguments")
